"""SMF specs from photonics.byu.edu/FiberOpticConnectors.parts/images/smf28.pdf.

MFD:

- 10.4 for Cband
- 9.2 for Oband

"""
from __future__ import annotations

import hashlib
import pathlib
import shlex
import shutil
import subprocess
import time
from typing import Dict, Optional, Tuple

import matplotlib.pyplot as plt
import meep as mp
import numpy as np
import omegaconf

from gdsfactory.config import logger, sparameters_path
from gdsfactory.serialization import clean_value_json, clean_value_name
from gdsfactory.simulation.gmeep.get_simulation_grating_fiber import (
    get_simulation_grating_fiber,
)
from gdsfactory.simulation.gmeep.write_sparameters_meep_mpi import _python
from gdsfactory.types import PathType

nm = 1e-3
nSi = 3.48
nSiO2 = 1.44

Floats = Tuple[float, ...]


def fiber_ncore(fiber_numerical_aperture, fiber_nclad):
    return (fiber_numerical_aperture**2 + fiber_nclad**2) ** 0.5


def write_sparameters_grating(
    plot: bool = False,
    plot_contour: bool = False,
    animate: bool = False,
    overwrite: bool = False,
    dirpath: Optional[PathType] = sparameters_path,
    decay_by: float = 1e-3,
    verbosity: int = 0,
    **settings,
) -> np.ndarray:
    """Write grating coupler with fiber Sparameters.

    Args:
        plot: plot simulation (do not run).
        plot_contour: show contours.
        animate: create animation.
        overwrite: overwrites simulation if found.
        dirpath: directory path.
        decay_by: field decay to stop simulation.
        verbosity: print messages.
        ncores: number of cores.

    Keyword Args:
        period: fiber grating period in um.
        fill_factor: fraction of the grating period filled with the grating material.
        n_periods: number of periods.
        widths: Optional list of widths. Overrides period, fill_factor, n_periods.
        gaps: Optional list of gaps. Overrides period, fill_factor, n_periods.
        fiber_angle_deg: fiber angle in degrees.
        fiber_xposition: xposition.
        fiber_core_diameter: fiber diameter.
        fiber_numerical_aperture: NA.
        fiber_nclad: fiber cladding index.
        nwg: waveguide index.
        nslab: slab refractive index.
        nclad: top cladding index.
        nbox: box index bottom.
        nsubstrate: index substrate.
        pml_thickness: pml_thickness (um).
        substrate_thickness: substrate_thickness (um).
        box_thickness: thickness for bottom cladding (um).
        wg_thickness: wg_thickness (um).
        slab_thickness: slab thickness (um). etch_depth=wg_thickness-slab_thickness.
        top_clad_thickness: thickness of the top cladding.
        air_gap_thickness: air gap thickness.
        fiber_thickness: fiber_thickness.
        resolution: resolution pixels/um.
        wavelength_start: min wavelength (um).
        wavelength_stop: max wavelength (um).
        wavelength_points: wavelength points.
        eps_averaging: epsilon averaging.
        fiber_port_y_offset_from_air: y_offset from fiber to air (um).
        waveguide_port_x_offset_from_grating_start: in um.
        fiber_port_x_size: in um.
        xmargin: margin from PML to grating end in um.

    .. code::

                 fiber_xposition
                  |
             fiber_core_diameter
          /     /  /     /       |
         /     /  /     /        | fiber_thickness
        /     /  /     /    _ _ _| _ _ _ _ _ _  _
                                 |
                                 | air_gap_thickness
                            _ _ _| _ _ _ _ _ _  _
                                 |
                nclad            | top_clad_thickness
                            _ _ _| _ _ _ _ _ _  _
             _|-|_|-|_|-|___     |              _| etch_depth
              ncore        |     |core_thickness
             ______________|_ _ _|_ _ _ _ _ _ _ _
                                 |
              nbox               |box_thickness
             ______________ _ _ _|_ _ _ _ _ _ _ _
                                 |
              nsubstrate         |substrate_thickness
             ______________ _ _ _|

    """
    mp.verbosity(verbosity)

    settings = clean_value_json(settings)
    settings_string = clean_value_name(settings)
    settings_hash = hashlib.md5(settings_string.encode()).hexdigest()[:8]

    filename = f"fiber_{settings_hash}.yml"
    dirpath = dirpath or pathlib.Path(__file__).parent / "data"
    dirpath = pathlib.Path(dirpath)
    dirpath.mkdir(exist_ok=True, parents=True)
    filepath = dirpath / filename
    filepath_npz = filepath.with_suffix(".npz")
    filepath_mp4 = filepath.with_suffix(".mp4")

    if filepath_npz.exists() and not overwrite and not plot:
        logger.info(f"sparameters loaded from {str(filepath_npz)!r}")
        return np.load(filepath_npz)

    sim_dict = get_simulation_grating_fiber(**settings)
    sim = sim_dict["sim"]
    freqs = sim_dict["freqs"]
    start = time.time()

    if plot or plot_contour:
        eps_parameters = dict(contour=True) if plot_contour else None
        sim.plot2D(eps_parameters=eps_parameters)
        plt.show()
        return

    termination = [mp.stop_when_energy_decayed(dt=50, decay_by=decay_by)]

    if animate:
        # Run while saving fields
        # sim.use_output_directory()
        animate = mp.Animate2D(
            sim,
            fields=mp.Ez,
            realtime=False,
            normalize=True,
            eps_parameters={"contour": True},
            field_parameters={
                "alpha": 0.8,
                "cmap": "RdBu",
                "interpolation": "none",
            },
            boundary_parameters={
                "hatch": "o",
                "linewidth": 1.5,
                "facecolor": "y",
                "edgecolor": "b",
                "alpha": 0.3,
            },
        )

        sim.run(mp.at_every(1, animate), until_after_sources=termination)
        animate.to_mp4(15, filepath_mp4)

    else:
        sim.run(until_after_sources=termination)

    # Extract mode information
    waveguide_monitor = sim_dict["waveguide_monitor"]
    waveguide_port_direction = sim_dict["waveguide_port_direction"]
    fiber_monitor = sim_dict["fiber_monitor"]
    fiber_angle_deg = sim_dict["fiber_angle_deg"]
    fcen = sim_dict["fcen"]
    wavelengths = 1 / freqs

    waveguide_mode = sim.get_eigenmode_coefficients(
        waveguide_monitor,
        [1],
        eig_parity=mp.ODD_Z,
        direction=waveguide_port_direction,
    )
    fiber_mode = sim.get_eigenmode_coefficients(
        fiber_monitor,
        [1],
        direction=mp.NO_DIRECTION,
        eig_parity=mp.ODD_Z,
        kpoint_func=lambda f, n: mp.Vector3(0, fcen * 1.45, 0).rotate(
            mp.Vector3(z=1), -1 * np.radians(fiber_angle_deg)
        ),  # Hardcoded index for now, pull from simulation eventually
    )
    end = time.time()

    a1 = waveguide_mode.alpha[:, :, 0].flatten()  # forward wave
    b1 = waveguide_mode.alpha[:, :, 1].flatten()  # backward wave

    # Since waveguide port is oblique, figure out forward and backward direction
    kdom_fiber = fiber_mode.kdom[0]
    idx = 1 - (kdom_fiber.y > 0) * 1

    a2 = fiber_mode.alpha[:, :, idx].flatten()  # forward wave
    # b2 = fiber_mode.alpha[:, :, 1 - idx].flatten()  # backward wave

    s11 = np.squeeze(b1 / a1)
    s12 = np.squeeze(a2 / a1)
    s22 = s11.copy()
    s21 = s12.copy()

    simulation = dict(
        settings=settings,
        compute_time_seconds=end - start,
        compute_time_minutes=(end - start) / 60,
    )
    filepath.write_text(omegaconf.OmegaConf.to_yaml(simulation))

    sp = {
        "o1@0,o1@0": s11,
        "o1@0,o2@0": s12,
        "o2@0,o1@0": s21,
        "o2@0,o2@0": s22,
        "wavelengths": wavelengths,
    }
    np.savez_compressed(filepath_npz, **sp)
    return sp


def write_sparameters_grating_mpi(
    instance: Dict,
    cores: int = 2,
    temp_dir: Optional[str] = None,
    temp_file_str: str = "write_sparameters_meep_mpi",
    verbosity: bool = False,
):
    """Write grating coupler Sparameters using multiple cores.

    Given a Dict of write_sparameters_meep keyword arguments (the "instance"),
    launches a parallel simulation on `cores` cores
    Returns the subprocess Popen object

    Args:
        instances: keys are parameters names of write_sparameters_meep,
            and entries the values.
        cores (int): number of processors.
        temp_dir (FilePath): temporary directory to hold simulation files.
        temp_file_str (str): names of temporary files in temp_dir.
        verbosity (bool): progress messages.

    """
    # Save the component object to simulation for later retrieval
    temp_dir = temp_dir or pathlib.Path(__file__).parent / "temp"
    temp_dir = pathlib.Path(temp_dir)
    temp_dir.mkdir(exist_ok=True, parents=True)
    filepath = temp_dir / temp_file_str

    # Add parallelism info
    instance["ncores"] = cores

    # Write execution file
    script_lines = [
        "from gdsfactory.simulation.gmeep.write_sparameters_grating import write_sparameters_grating\n\n",
        'if __name__ == "__main__":\n\n',
        "\twrite_sparameters_grating(\n",
    ]
    for key in instance.keys():
        if isinstance(instance[key], str):
            parameter = f'"{instance[key]}"'
        else:
            parameter = instance[key]
        script_lines.append(f"\t\t{key} = {parameter},\n")
    script_lines.append("\t)")
    script_file = filepath.with_suffix(".py")
    with open(script_file, "w") as script_file_obj:
        script_file_obj.writelines(script_lines)
    # Exec string
    command = f"mpirun -np {cores} {_python()} {script_file}"

    # Launch simulation
    if verbosity:
        print(f"Launching: {command}")
    return subprocess.Popen(
        shlex.split(command),
        shell=False,
        stdin=None,
        stdout=None,
        stderr=None,
    )


def write_sparameters_grating_batch(
    instances,
    cores_per_instance: int = 2,
    total_cores: int = 4,
    temp_dir: Optional[str] = None,
    delete_temp_files: bool = False,
    verbosity: bool = False,
) -> None:
    """Write grating coupler Sparameters using multiple cores in batches of simulations.

    Given a tuple of write_sparameters_meep keyword arguments (instances)
    launches parallel simulations each simulation is assigned "cores_per_instance" cores
    Assumes total of "total_cores" if cores_per_instance * len(instances) > total_cores
    then the overflow will be performed serially

    Args:
        instances: list of Dicts. The keys must be parameters names of write_sparameters_meep,
            and entries the values.
        cores_per_instance: number of processors to assign to each instance.
        total_cores: total number of cores to use.
        temp_dir: temporary directory to hold simulation files.
        delete_temp_file: whether to delete temp_dir when done.
        verbosity: show progress messages.

    """
    # Save the component object to simulation for later retrieval
    temp_dir = temp_dir or pathlib.Path(__file__).parent / "temp"
    temp_dir = pathlib.Path(temp_dir)
    temp_dir.mkdir(exist_ok=True, parents=True)

    # Setup pools
    num_pools = int(np.ceil(cores_per_instance * len(instances) / total_cores))
    instances_per_pool = int(np.floor(total_cores / cores_per_instance))
    num_tasks = len(instances)

    if verbosity:
        print(f"Running parallel simulations over {num_tasks} instances")
        print(
            f"Using a total of {total_cores} cores with {cores_per_instance} cores per instance"
        )
        print(
            f"Tasks split amongst {num_pools} pools with up to {instances_per_pool} instances each."
        )

    i = 0
    # For each pool
    for j in range(num_pools):
        processes = []
        # For instance in the pool
        for k in range(instances_per_pool):
            # Flag to catch nonfull pools
            if i >= num_tasks:
                continue
            if verbosity:
                print(f"Task {k} of pool {j} is instance {i}")
            # Obtain current instance
            instance = instances[i]

            process = write_sparameters_grating_batch(
                instances=instance,
                cores_per_instance=cores_per_instance,
                temp_dir=temp_dir,
                verbosity=verbosity,
            )
            processes.append(process)

            # Increment task number
            i += 1

        # Wait for pool to end
        for process in processes:
            process.wait()

    if delete_temp_files:
        shutil.rmtree(temp_dir)


if __name__ == "__main__":
    # fiber_numerical_aperture = float(np.sqrt(1.44427 ** 2 - 1.43482 ** 2))
    # instance1 = dict(
    #     period=0.66,
    #     fill_factor=0.5,
    #     n_periods=50,
    #     etch_depth=70 * nm,
    #     fiber_angle_deg=10.0,
    #     fiber_xposition=0.0,
    #     fiber_core_diameter=9,
    #     fiber_numerical_aperture=fiber_numerical_aperture,
    #     fiber_nclad=1.43482,
    #     ncore=3.47,
    #     nclad=1.44,
    #     nbox=1.44,
    #     nsubstrate=3.47,
    #     pml_thickness=1.0,
    #     substrate_thickness=1.0,
    #     box_thickness=2.0,
    #     core_thickness=220 * nm,
    #     top_clad_thickness=2.0,
    #     air_gap_thickness=1.0,
    #     fiber_thickness=2.0,
    #     res=20,  # pixels/um
    #     wavelength_min=1.4,
    #     wavelength_max=1.7,
    #     wavelength_points=150,
    #     fiber_port_y_offset_from_air=1,
    #     waveguide_port_x_offset_from_grating_start=10,
    #     overwrite=True,
    #     verbosity=0,
    #     decay_by=1e-3,
    # )
    # instance2 = instance1.copy()
    # instance2["period"] = 0.5
    # write_sparameters_meep_batch(
    #     instances=[instance1, instance2],
    #     cores_per_instance=4,
    #     total_cores=8,
    #     verbosity=True,
    #     delete_temp_files=True,
    # )

    from gdsfactory.simulation.plot import plot_sparameters

    sp = write_sparameters_grating(fiber_angle_deg=15)
    plot_sparameters(sp)



from __future__ import annotations

import pytest

import gdsfactory as gf
from gdsfactory.tech import LAYER


@pytest.mark.parametrize("optical_routing_type", [0, 1])
def test_add_pins_with_routes(optical_routing_type) -> gf.Component:
    """Add grating couplers to a straight ensure that all the routes have
    pins."""
    c = gf.components.straight(length=11.0)
    gc = gf.components.grating_coupler_elliptical_te()
    cc = gf.routing.add_fiber_single(
        component=c,
        grating_coupler=[gc, gf.components.grating_coupler_tm],
        optical_routing_type=optical_routing_type,
    )
    pins_component = cc.extract(layers=(LAYER.PORT,))
    assert len(pins_component.paths) == 12, len(pins_component.paths)
    return pins_component


def test_add_pins() -> None:
    """Ensure that all the waveguide has 2 pins."""
    c = gf.components.straight(length=11.0)
    pins_component = c.extract(layers=(LAYER.PORT,))
    assert len(pins_component.paths) == 2, len(pins_component.paths)


if __name__ == "__main__":
    # test_add_pins()
    c = test_add_pins_with_routes(0)
    c.show(show_ports=True)

    # test_add_pins_with_routes(1)

    # c = straight_nc()
    # gc = gf.components.grating_coupler_elliptical_te(wg_width=WIDTH_NITRIDE_CBAND)
    # cc = gf.routing.add_fiber_single(
    #     component=c,
    #     grating_coupler=gc,
    #     straight=straight_nc,
    #     optical_routing_type=1,
    # )
    # cc.show(show_ports=True)
    # pins_component = cc.extract(layers=(LAYER.PIN,))
    # print(len(pins_component.polygons))

    # c = gf.components.straight(length=11.0)
    # c2 = c.extract(layers=(LAYER.PORT,))
    # print(c2.paths)
    # print(c2.polygons)
    # c2.show()

    # c = gf.components.straight(length=11.0)
    # gc = gf.components.grating_coupler_elliptical_te()
    # cc = gf.routing.add_fiber_single(
    #     c,
    #     grating_coupler=[gc, gf.components.grating_coupler_tm],
    #     optical_routing_type=0,
    # )
    # cc.show()



from __future__ import annotations

from typing import Optional

import numpy as np

import gdsfactory as gf
from gdsfactory.component import Component
from gdsfactory.components.rectangle import rectangle
from gdsfactory.components.taper import taper
from gdsfactory.types import ComponentSpec, CrossSectionSpec, LayerSpec


@gf.cell
def grating_coupler_rectangular(
    n_periods: int = 20,
    period: float = 0.75,
    fill_factor: float = 0.5,
    width_grating: float = 11.0,
    length_taper: float = 150.0,
    polarization: str = "te",
    wavelength: float = 1.55,
    taper: ComponentSpec = taper,
    layer_slab: Optional[LayerSpec] = "SLAB150",
    fiber_marker_layer: LayerSpec = "TE",
    slab_xmin: float = -1.0,
    slab_offset: float = 1.0,
    cross_section: CrossSectionSpec = "strip",
    **kwargs,
) -> Component:
    r"""Grating coupler with rectangular shapes (not elliptical).

    Needs longer taper than elliptical.
    Grating teeth are straight.
    For a focusing grating take a look at grating_coupler_elliptical.

    Args:
        n_periods: number of grating teeth.
        period: grating pitch.
        fill_factor: ratio of grating width vs gap.
        width_grating: 11.
        length_taper: 150.
        wg_width: input waveguide width.
        layer: for grating teeth.
        polarization: 'te' or 'tm'.
        wavelength: in um.
        taper: function.
        layer_slab: layer that protects the slab under the grating.
        slab_xmin: where 0 is at the start of the taper.
        slab_offset: from edge of grating to edge of the slab.
        cross_section: for input waveguide port.
        kwargs: cross_section settings.

    .. code::

        side view
                      fiber

                   /  /  /  /
                  /  /  /  /

                _|-|_|-|_|-|___ layer
                   layer_slab |
            o1  ______________|


        top view     _________
                    /| | | | |
                   / | | | | |
                  /taper_angle
                 /_ _| | | | |
        wg_width |   | | | | |
                 \   | | | | |
                  \  | | | | |
                   \ | | | | |
                    \|_|_|_|_|
                 <-->
                taper_length
    """
    xs = gf.get_cross_section(cross_section, **kwargs)
    wg_width = xs.width
    layer = xs.layer

    c = Component()
    taper_ref = c << gf.get_component(
        taper,
        length=length_taper,
        width2=width_grating,
        width1=wg_width,
        layer=layer,
    )

    c.add_port(port=taper_ref.ports["o1"], name="o1")
    x0 = taper_ref.xmax

    for i in range(n_periods):
        xsize = gf.snap.snap_to_grid(period * fill_factor)
        cgrating = c.add_ref(
            rectangle(size=(xsize, width_grating), layer=layer, port_type=None)
        )
        cgrating.xmin = gf.snap.snap_to_grid(x0 + i * period)
        cgrating.y = 0

    xport = np.round((x0 + cgrating.x) / 2, 3)

    port_type = f"vertical_{polarization.lower()}"
    c.add_port(
        name=port_type,
        port_type=port_type,
        center=(xport, 0),
        orientation=0,
        width=width_grating,
        layer=fiber_marker_layer,
    )
    c.info["polarization"] = polarization
    c.info["wavelength"] = wavelength
    gf.asserts.grating_coupler(c)

    if layer_slab:
        slab_xmin += length_taper
        slab_xsize = cgrating.xmax + slab_offset
        slab_ysize = c.ysize + 2 * slab_offset
        yslab = slab_ysize / 2
        c.add_polygon(
            [
                (slab_xmin, yslab),
                (slab_xsize, yslab),
                (slab_xsize, -yslab),
                (slab_xmin, -yslab),
            ],
            layer_slab,
        )
    if xs.add_bbox:
        c = xs.add_bbox(c)
    if xs.add_pins:
        c = xs.add_pins(c)
    return c


if __name__ == "__main__":
    # c = grating_coupler_rectangular(name='gcu', partial_etch=True)
    # c = grating_coupler_rectangular()
    c = gf.routing.add_fiber_array(grating_coupler=grating_coupler_rectangular)
    print(c.ports)
    c.show(show_ports=True)



from __future__ import annotations

from typing import Optional

import gdsfactory as gf
from gdsfactory.component import Component
from gdsfactory.components.taper_cross_section import taper_cross_section
from gdsfactory.types import CrossSectionSpec, LayerSpecs


@gf.cell
def terminator(
    length: Optional[float] = 50,
    cross_section_input: CrossSectionSpec = "strip",
    cross_section_tip: Optional[CrossSectionSpec] = None,
    tapered_width: float = 0.2,
    doping_layers: LayerSpecs = ("NPP",),
) -> gf.Component:
    """Returns doped taper to terminate waveguides.

    Args:
        length: distance between input and narrow tapered end.
        cross_section_input: input cross-section.
        cross_section_tip: cross-section at the end of the termination.
        tapered_width: width of the default cross-section at the end of the termination.
            Only used if cross_section_tip is not None.
        doping_layers: doping layers to superimpose on the taper. Default N++.
    """
    c = Component()

    cross_section_tip = (
        cross_section_tip
        if cross_section_tip
        else gf.get_cross_section(cross_section_input, width=tapered_width)
    )

    taper = c << gf.get_component(
        taper_cross_section,
        length=length,
        cross_section1=cross_section_input,
        cross_section2=cross_section_tip,
    )

    for layer in doping_layers:
        c << gf.components.bbox(bbox=taper.bbox, layer=layer)

    c.add_port(name="o1", port=taper.ports["o1"])
    return c


if __name__ == "__main__":
    c = terminator(cross_section_input=gf.partial(gf.cross_section.strip, width=10))
    c.show(show_ports=True)



from collections import OrderedDict

from shapely.geometry import LineString, MultiLineString, MultiPolygon, Polygon
from shapely.ops import linemerge, split

from gdsfactory.simulation.gmsh.parse_gds import tile_shapes


def break_line(line, other_line):
    intersections = line.intersection(other_line)
    if not intersections.is_empty:
        for intersection in (
            intersections.geoms if hasattr(intersections, "geoms") else [intersections]
        ):
            if intersection.type == "Point":
                line = linemerge(split(line, intersection))
            else:
                new_coords_start, new_coords_end = intersection.boundary.geoms
                line = linemerge(split(line, new_coords_start))
                line = linemerge(split(line, new_coords_end))

    return line


def break_geometry(shapes_dict: OrderedDict):
    """Break up lines and polygon edges so that plane is tiled with no partially overlapping line segments.

    Args:
        shapes_dict: arbitrary dict of shapely polygons and lines, with ordering setting mesh priority

    Returns:
        polygons_broken_dict: dict of shapely polygons, with smallest number of individual vertices
        lines_broken_dict: dict of shapely lines, with smallest number of individual vertices
    """
    # Break up shapes in order so that plane is tiled with non-overlapping layers
    shapes_tiled_dict = tile_shapes(shapes_dict)

    polygons_broken_dict = OrderedDict()
    lines_broken_dict = OrderedDict()
    for _first_index, (first_name, _init_first_shapes) in enumerate(
        shapes_dict.items()
    ):
        first_shapes = shapes_tiled_dict[first_name]
        broken_shapes = []
        for first_shape in (
            first_shapes.geoms if hasattr(first_shapes, "geoms") else [first_shapes]
        ):
            # First line exterior
            first_exterior_line = (
                LineString(first_shape.exterior)
                if first_shape.type == "Polygon"
                else first_shape
            )
            for second_name, second_shapes in shapes_dict.items():
                # Do not compare to itself
                if second_name != first_name:
                    second_shapes = shapes_tiled_dict[second_name]
                    for second_shape in (
                        second_shapes.geoms
                        if hasattr(second_shapes, "geoms")
                        else [second_shapes]
                    ):
                        # Second line exterior
                        second_exterior_line = (
                            LineString(second_shape.exterior)
                            if second_shape.type == "Polygon"
                            else second_shape
                        )
                        first_exterior_line = break_line(
                            first_exterior_line, second_exterior_line
                        )
                        # Second line interiors
                        for second_interior_line in (
                            second_shape.interiors
                            if second_shape.type == "Polygon"
                            else []
                        ):
                            second_interior_line = LineString(second_interior_line)
                            first_exterior_line = break_line(
                                first_exterior_line, second_interior_line
                            )
            # First line interiors
            if first_shape.type in ["Polygon", "MultiPolygon"]:
                first_shape_interiors = []
                for first_interior_line in first_shape.interiors:
                    first_interior_line = LineString(first_interior_line)
                    for second_name, second_shapes in shapes_dict.items():
                        if second_name != first_name:
                            second_shapes = shapes_tiled_dict[second_name]
                            for second_shape in (
                                second_shapes.geoms
                                if hasattr(second_shapes, "geoms")
                                else [second_shapes]
                            ):
                                # Exterior
                                second_exterior_line = (
                                    LineString(second_shape.exterior)
                                    if second_shape.type == "Polygon"
                                    else second_shape
                                )
                                first_interior_line = break_line(
                                    first_interior_line, second_exterior_line
                                )
                                # Interiors
                                for second_interior_line in (
                                    second_shape.interiors
                                    if second_shape.type == "Polygon"
                                    else []
                                ):
                                    second_interior_line = LineString(
                                        second_interior_line
                                    )
                                    first_interior_line = break_line(
                                        first_interior_line, second_interior_line
                                    )
                    first_shape_interiors.append(first_interior_line)
            if first_shape.type in ["Polygon", "MultiPolygon"]:
                broken_shapes.append(
                    Polygon(first_exterior_line, holes=first_shape_interiors)
                )
            else:
                broken_shapes.append(LineString(first_exterior_line))
        if broken_shapes:
            if first_shape.type in ["Polygon", "MultiPolygon"]:
                polygons_broken_dict[first_name] = (
                    MultiPolygon(broken_shapes)
                    if len(broken_shapes) > 1
                    else broken_shapes[0]
                )
            else:
                lines_broken_dict[first_name] = (
                    MultiLineString(broken_shapes)
                    if len(broken_shapes) > 1
                    else broken_shapes[0]
                )

    return polygons_broken_dict, lines_broken_dict



from __future__ import annotations

import numpy as np

import gdsfactory as gf


@gf.cell
def rectangles3():
    c = gf.Component()
    c << gf.c.rectangle(size=(4, 4), layer=(1, 0))
    c << gf.c.rectangle(size=(4, 4), layer=(2, 0))
    c.distribute()
    return c


def test_remove_layers():
    c0 = rectangles3()
    assert np.isclose(c0.area(), 16.0 * 2)

    c1 = c0.remove_layers([(1, 0)])
    assert np.isclose(c1.area(), 16.0)


if __name__ == "__main__":
    c0 = rectangles3()
    assert np.isclose(c0.area(), 16.0 * 2)

    # c1 = c0.remove_layers([(1, 0)])
    # assert np.isclose(c1.area(), 16.0)
    c0._cell.filter(
        [(1, 0)],
    )
    c0.show()



"""Merge mask metadata with test labels to return test_metadata."""
from __future__ import annotations

import warnings
from pathlib import Path
from typing import Any, Dict, List, Optional

from omegaconf import DictConfig, OmegaConf

from gdsfactory.config import logger
from gdsfactory.types import PathType


def parse_csv_data(
    csv_labels_path: Path, ignore_prefix: str = "METR_"
) -> List[List[str]]:
    """Returns CSV labels as a list of strings."""
    with open(csv_labels_path) as f:
        # Get all lines
        lines = [line.replace("\n", "") for line in f.readlines()]

        # Ignore labels for metrology structures
        lines = [line for line in lines if not line.startswith(ignore_prefix)]

        # Split lines in fields
        lines = [line.split(",") for line in lines]

        lines = [[s.strip() for s in split if s.strip()] for split in lines]

        # Remove empty lines
        lines = [line for line in lines if line]
    return lines


def get_cell_from_label(label: str) -> str:
    """Get cell name from the label (cell_name is in parenthesis)."""
    try:
        cell_name = label.split("(")[1].split(")")[0]
    except IndexError as error:
        raise ValueError(f"{label!r} needs (cell name) between parenthesis") from error

    if cell_name.startswith("loopback"):
        cell_name = "_".join(cell_name.split("_")[1:])
    return cell_name


def merge_test_metadata(
    labels_path: PathType,
    mask_metadata: Dict[str, Any],
    labels_prefix: str = "opt",
    get_cell_from_string=get_cell_from_label,
    filepath: Optional[PathType] = None,
) -> DictConfig:
    """Returns a test metadata dict config of labeled cells by merging GDS \
    labels in CSV and YAML mask metadata.

    Args:
        labels_path: for test labels in CSV.
        mask_metadata: dict with test metadata.
        labels_prefix: only select labels with a text prefix.
        get_cell_from_string: returns label string.
        filepath: Optional path to write test metadata.

    .. code::

        CSV labels  -------
                          |--> merge_test_metadata dict
                          |
        YAML metadata  ----

    """
    labels_path = Path(labels_path)

    if not labels_path.exists():
        raise FileNotFoundError(f"missing CSV labels {labels_path!r}")

    labels_list = parse_csv_data(labels_path)
    cells_metadata = mask_metadata.get("cells", {})

    test_metadata = DictConfig({})

    for label, x, y in labels_list:
        cell = get_cell_from_string(label)

        if cell in cells_metadata:
            test_metadata[cell] = cells_metadata[cell]
            test_metadata[cell].label = dict(x=float(x), y=float(y), text=label)
        else:
            logger.error(f"missing cell metadata for {cell!r}")
            warnings.warn(f"missing cell metadata for {cell!r}")

    if filepath:
        filepath = Path(filepath)
        filepath.write_text(OmegaConf.to_yaml(test_metadata))

    return test_metadata


if __name__ == "__main__":
    # from gdsfactory import PATH

    # labels_path = (
    #    PATH.samples_path / "mask_pack" / "build" / "mask" / "sample_mask.csv"
    # )
    # mask_metadata_path = labels_path.with_suffix(".yml")
    # mask_metadata = OmegaConf.load(mask_metadata_path)
    # d = merge_test_metadata(labels_path=labels_path, mask_metadata=mask_metadata)
    # print(d)
    print(get_cell_from_label("opt_te1550_demo"))



from __future__ import annotations

import pandas as pd
import yaml

import gdsfactory as gf


def test_import_ports_markers_labels(data_regression) -> gf.Component:
    """Make sure you can import the ports"""
    c = gf.components.straight(
        decorator=gf.add_pins.add_pins_siepic, cross_section="strip_no_pins"
    )
    c1 = gf.labels.add_label_yaml(
        gf.functions.rotate(gf.routing.add_fiber_single(c), angle=90)
    )
    gdspath = c1.write_gds()
    csvpath = gf.labels.write_labels.write_labels_gdstk(
        gdspath, prefix="component_name"
    )
    labels = pd.read_csv(csvpath)

    settings = yaml.safe_load(labels.columns[0])
    data_regression.check(settings)


if __name__ == "__main__":
    c = gf.components.straight(
        decorator=gf.add_pins.add_pins_siepic, cross_section="strip_no_pins"
    )
    c1 = gf.labels.add_label_yaml(
        gf.functions.rotate(gf.routing.add_fiber_single(c), angle=90)
    )
    gdspath = c1.write_gds()
    csvpath = gf.labels.write_labels.write_labels_gdstk(
        gdspath, prefix="component_name"
    )
    labels = pd.read_csv(csvpath)
    settings = yaml.safe_load(labels.columns[0])
    c1.show()
